////////////////////////////////////////////////////////////////////// 
// extras/pedstats.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday June 11, 2003
// 
 
#include "pedstats.h"
#include "pedstats_pdf.h"

String pedName = "pedstats.ped";
String dataName = "pedstats.dat";
String ibdName = "pedstats.ibd";
String pdfName = "pedstats.pdf";

Pedigree ped;
IBDTable ibd;

bool hardy = false, hardyAll = false, hardyFounders = false;
bool nocheck = false;
bool pairs = false;
bool trimPedigree = false;
bool rewritePedigree = false;
bool writePDF = false;

int main(int argc, char * argv[])
   {
   printf("Pedigree Statistics\n(c) 1999-2003 Goncalo Abecasis\n\n");

   ReadParameters(argc, argv);

   ped.Prepare(dataName);
   ped.Load(pedName);

   if (trimPedigree) ped.Trim();

   ibd.Load(ped, ibdName);

   FamilyStatistics();
   TraitStatistics();
   CovarStatistics();
   AffectionStatistics();

   if (!nocheck) ped.InheritanceCheck();

   MarkerStatistics();

   if (hardy || hardyAll) HWStatistics();
   if (hardy || hardyFounders) HWStatistics(true);

   if (writePDF)
      {
      PDF pdf;
      pdf.OpenFile(pdfName);
      pdf.page.SetSize(psLetterR);
      GraphFamilyStatistics(pdf);
      GraphTraitStatistics(pdf);
      GraphCovariateStatistics(pdf);
      GraphAffectionStatistics(pdf);
      GraphMarkerStatistics(pdf);
      }

   return 0;
   };

int Generation(Person * p)
   {
   if (p->isFounder())
      return 0;

   int father = Generation(p->father);
   int mother = Generation(p->mother);

   return (mother > father ? mother : father) + 1;
   }

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Pedigree File")
      LONG_PARAMETER("ignoreMendelianErrors", &nocheck)
      LONG_PARAMETER("chromosomeX", &Pedigree::chromosomeX)
      LONG_PARAMETER("trim", &trimPedigree)
   LONG_PARAMETER_GROUP("Genotype Checks")
      LONG_PARAMETER("hardyWeinberg", &hardy)
      LONG_PARAMETER("checkFounders", &hardyFounders)
      LONG_PARAMETER("checkAll", &hardyAll)
   LONG_PARAMETER_GROUP("Output Options")
      LONG_PARAMETER("pairStatistics", &pairs)
      LONG_PARAMETER("rewritePedigree", &rewritePedigree)
      LONG_PARAMETER("pdf", &writePDF)
END_LONG_PARAMETERS();

void ReadParameters(int argc, char * argv[])
   {
   ParameterList pl;

   pl.Add(new StringParameter('p', "Pedigree File", pedName));
   pl.Add(new StringParameter('d', "Data File", dataName));
   pl.Add(new StringParameter('i', "IBD File", ibdName));
   pl.Add(new StringParameter('a', "Adobe PDF File", pdfName));
   pl.Add(new StringParameter('x', "Missing Value Code", Pedigree::missing));
   pl.Add(new LongParameters("Additional Options", longParameters));

   pl.Read(argc, argv);
   pl.Status();
   }

void FamilyStatistics()
   {
   if (ped.familyCount == 0)
      {
      printf("WARNING - Pedigree is empty");
      return;
      }

   int families = 0, founders = 0, nonfounders = 0;
   int maxgen = 0, mingen = 9999;
   int maxcount, mincount;
   double avecount = 0.0, avegen = 0.0;

   maxcount = mincount = ped.families[0]->count;

   IntArray generations(ped.familyCount);

   for (int f = 0; f < ped.familyCount; f++)
      {
      families++;
      founders    += ped.families[f]->founders;
      nonfounders += ped.families[f]->count - ped.families[f]->founders;

      maxcount = max(ped.families[f]->count, maxcount);
      mincount = min(ped.families[f]->count, mincount);
      avecount += ped.families[f]->count;

      generations[f] = 0;
      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         generations[f] = max(Generation(ped.persons[i]) + 1, generations[f]);

      maxgen = max(generations[f], maxgen);
      mingen = min(generations[f], mingen);
      avegen += generations[f];
      }

   avegen /= (double) max(families, 1);
   avecount /= (double) max(families, 1);

   IntArray counts(maxcount + 1), gens(maxgen + 1);

   counts.Set(0);
   gens.Set(0);

   for (int f = 0; f < ped.familyCount; f++)
      {
      counts[ped.families[f]->count]++;
      gens[generations[f]]++;
      }

   IntArray idxcount(maxcount + 1), idxgen(maxgen + 1);

   MakeIndex(idxcount, counts);
   MakeIndex(idxgen, gens);

   double fdiv = max((double) families, 1.0) / (double) 100.0;

   printf("PEDIGREE STRUCTURE\n"
          "==================\n"
          "  Individuals: %d (%d founders, %d nonfounders)\n"
          "     Families: %d\n\n"
          " Family Sizes\n"
          "      Average: %.2f (%d to %d)\n"
          " Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n"
          "  Generations\n"
          "      Average: %.2f (%d to %d)\n"
          " Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n",
          founders + nonfounders, founders, nonfounders, families,
          avecount, mincount, maxcount,
          idxcount[0], (double) counts[idxcount[0]] / fdiv,
          idxcount[1], (double) counts[idxcount[1]] / fdiv,
          idxcount[2], (double) counts[idxcount[2]] / fdiv,
          avegen, mingen, maxgen,
          idxgen[0], (double) gens[idxgen[0]] / fdiv,
          idxgen[1], (double) gens[idxgen[1]] / fdiv,
          idxgen[2], (double) gens[idxgen[2]] / fdiv);

   printf("Checking family connectedness ... \n");
   int connected = true;
   for (int f = 0, groups; f < ped.familyCount; f++)
      if ( (groups = ped.families[f]->ConnectedGroups()) != 1 )
        printf("  Family %s (%d individuals) consists of %d disjoint groups\n",
          (const char *) ped.families[f]->famid, ped.families[f]->count, groups),
        connected = false;

   printf(connected ? "  All individuals in each family are connected.\n\n" :
                      "  Use the --rewritePedigree option to break families"
                      " into connected groups\n\n");

   if (rewritePedigree)
      {
      printf("Rewriting data file as [pedstats.dat]... ");
      ped.WriteDataFile("pedstats.dat");

      printf("Done!\n"
             "Rewriting data file as [pedstats.ped]... ");
      FILE * pedfile = fopen("pedstats.ped", "wt");

      IntArray groups;
      for (int f = 0; f < ped.familyCount; f++)
         {
         ped.families[f]->ConnectedGroups(&groups);

         for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
            ped.WritePerson(pedfile, i,
                            ped[i].famid + "_" + groups[ped[i].traverse]);
         }

      printf("Done!\n\n");
      fclose(pedfile);
      }

   if (pairs) RelativePairs();
   }

int trait_filter = 0;
bool FilterOnTrait(Person & p1, Person & p2)
   {
   return ((p1.traits[trait_filter] != _NAN_) &&
           (p2.traits[trait_filter] != _NAN_));
   }

void TraitStatistics()
   {
   if (Person::traitCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("QUANTITATIVE TRAIT STATISTICS\n"
          "=============================\n"
          "%15s %15s %15s %10s %10s\n",
          "", "[Phenotypes]", "[Founders]", "Mean", "Var");

   for (int t = 0; t < Person::traitCount; t++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     double mean = 0, var = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isPhenotyped(t) ) continue;
         mean    += ped[i].traits[t];
         var     += ped[i].traits[t] * ped[i].traits[t];
         tfcount += ped[i].isFounder();
         tcount  ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     var = (tcount <= 1) ? 0.0 :
            (tcount * var - mean * mean) /
           ((tcount) * (tcount - 1));
     mean /= max(tcount, 1);

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %10.3f %10.3f\n",
            (const char *) Person::traitNames[t],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            mean, var);
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );

   if (pairs)
      {
      for (trait_filter = 0; trait_filter <  Person::traitCount; trait_filter++)
         {
         if (trait_filter == 0)
            printf("%15s %12s %5s %9s %11s %10s %8s\n",
              "PairCounts", "ParentChild", "Sibs", "HalfSibs",
              "Grandparent", "Avuncular", "Cousins");

         printf("%15.15s ", (const char *) ped.traitNames[trait_filter]);
         RelativePairs(FilterOnTrait);
         }
      printf("\n");
      }
   }

int covariate_filter = 0;
bool FilterOnCovariate(Person & p1, Person & p2)
   {
   return ((p1.covariates[covariate_filter] != _NAN_) &&
           (p2.covariates[covariate_filter] != _NAN_));
   }

void CovarStatistics()
   {
   if (Person::covariateCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("COVARIATE STATISTICS\n"
          "====================\n"
          "%15s %15s %15s %10s %10s\n",
          "", "[Covariates]", "[Founders]", "Mean", "Var");

   for (int c = 0; c < Person::covariateCount; c++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     double mean = 0, var = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isControlled(c) ) continue;
         mean    += ped[i].covariates[c];
         var     += ped[i].covariates[c] * ped[i].covariates[c];
         tfcount += ped[i].isFounder();
         tcount  ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     var = (tcount <= 1) ? 0.0 :
            (tcount * var - mean * mean) /
           ((tcount) * (tcount - 1));
     mean /= max(tcount, 1);

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %10.3f %10.3f\n",
            (const char *) Person::covariateNames[c],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            mean, var);
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );

   if (pairs)
      {
      for (covariate_filter = 0; covariate_filter < Person::covariateCount; covariate_filter++)
         {
         if (covariate_filter == 0) printf("%15s %12s %5s %9s %11s %10s %8s\n",
              "PairCounts", "ParentChild", "Sibs", "HalfSibs",
              "Grandparent", "Avuncular", "Cousins");

         printf("%15.15s ", (const char *) ped.covariateNames[covariate_filter]);
         RelativePairs(FilterOnCovariate);
         }
      printf("\n");
      }

   }

int affection_filter = 0;
bool FilterOnAffection(Person & p1, Person & p2)
   {
   return ((p1.affections[affection_filter] == 2) &&
           (p2.affections[affection_filter] == 2));
   }

void AffectionStatistics()
   {
   if (Person::affectionCount == 0) return;

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   printf("AFFECTION STATISTICS\n"
          "====================\n"
          "%15s %15s %15s %10s\n",
          "", "[Diagnostics]", "[Founders]", "Prevalence");

   for (int a = 0; a < Person::affectionCount; a++)
     {
     int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
     int    affecteds = 0;

     for (int i = 0; i < ped.count; i++)
         {
         tpossible  ++;
         tfpossible += ped[i].isFounder();
         if ( !ped[i].isDiagnosed(a) ) continue;
         affecteds += ped[i].affections[a] == 2;
         tfcount   += ped[i].isFounder();
         tcount    ++;
         }

     possible  += tpossible;
     fpossible += tfpossible;
     count     += tcount;
     fcount    += tfcount;

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n",
            (const char *) Person::affectionNames[a],
            tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
            tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
            (double) affecteds / (double) max(tcount, 1) * 100 );
     }
   printf("%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );

   if (pairs)
      {
      for (affection_filter = 0; affection_filter < Person::affectionCount; affection_filter++)
         {
         if (affection_filter == 0)   printf("%15s %12s %5s %9s %11s %10s %8s\n",
              "PairCounts", "ParentChild", "Sibs", "HalfSibs",
              "Grandparent", "Avuncular", "Cousins");

         printf("%15.15s ", (const char *) ped.affectionNames[affection_filter]);
         RelativePairs(FilterOnAffection);
         }
      printf("\n");
      }
   }

void MarkerStatistics()
   {
   if (Person::markerCount == 0) return;

   printf("MARKER GENOTYPE STATISTICS\n"
          "==========================\n"
          "%15s %15s %15s %10s %8s\n",
          "", "[Genotypes]", "[Founders]", "Hetero", "IBD");

   int genotypes = 0, possible = 0, fgenotypes = 0, fpossible = 0,
       hetero = 0, maxHetero = 0;

   for (int m = 0; m < Person::markerCount; m++)
     {
     int mGenotypes = 0, mPossible = 0, mfGenotypes = 0, mfPossible = 0,
         mHetero = 0, mMaxHetero = 0, lastAllele = 0;

     for (int i = 0; i < ped.count; i++)
         {
         mPossible  ++;
         mfPossible += ped[i].isFounder();
         if ( !ped[i].isGenotyped(m) ) continue;
         mGenotypes  ++;
         mfGenotypes += ped[i].isFounder();

         lastAllele = max( max(ped[i].markers[m][0], ped[i].markers[m][1]),
                           lastAllele);

         if (ped.chromosomeX && ped[i].sex == SEX_MALE) continue;
         mHetero     += ped[i].markers[m].isHeterozygous();
         mMaxHetero  ++;
         }

     possible   += mPossible;
     fpossible  += mfPossible;
     fgenotypes += mfGenotypes;
     genotypes  += mGenotypes;
     hetero     += mHetero;
     maxHetero  += mMaxHetero;

     int ibdCount = 0;
     if (!ibd.isEmpty())
        for (int f = 0; f < ped.familyCount; f++)
            if (ibd.HaveFamily(m, ped.families[f]))
                ibdCount++;

     printf("%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%% %6d/%d families\n",
            (const char *) Person::markerNames[m],
            mGenotypes, (double) mGenotypes / (double) max(mPossible, 1) * 100,
            mfGenotypes, (double)mfGenotypes / (double)max(mfPossible, 1) * 100,
            (double) mHetero / (double) max(mMaxHetero, 1) * 100,
            ibdCount, ped.familyCount);
     }

     printf("%15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n%s\n", "Total",
            genotypes, (double) genotypes / (double) max(possible, 1) * 100,
            fgenotypes, (double) fgenotypes / (double) max(fpossible, 1) * 100,
            (double) hetero / (double) max(maxHetero, 1) * 100,
            ped.chromosomeX ?
            "\nNOTE: Heterozygosity calculation excludes males\n" : "");
   }

void MakeIndex(IntArray & index, IntArray & counts)
   {
   index.Dimension(counts.Length());
   index.SetSequence(0, 1);

   // Insertion sort should be OK, for N < 20
   for (int i, j = 1; j < counts.Length(); j++)
      {
      int tmp = counts[j];    // index[j] = j
      for (i = j - 1; (i >= 0) && (counts[index[i]] < tmp); i--)
         index[i + 1] = index[i];
      index[i + 1] = j;
      }

   // Make sure we have at least three items
   for (int i = counts.Length() + 1; i < 4; i++)
      {
      counts.Dimension(i);
      index.Dimension(i);

      counts[i - 1] = 0;
      index[i - 1] = i - 1;
      }
   }

void HWStatistics(bool foundersOnly)
   {
   if (ped.markerCount == 0) return;

   bool fail = false;
   bool note = false;

   if (ped.chromosomeX)
      printf("HARDY-WEINBERG CHECK (AMONG %sFEMALES)\n"
             "============================%s========\n",
             foundersOnly ? "FOUNDER " : "",
             foundersOnly ? "========" : "");
   else
      printf("HARDY-WEINBERG CHECK%s\n"
             "====================%s\n",
             foundersOnly ? " AMONG FOUNDERS" : "",
             foundersOnly ? "===============" : "");

   Matrix   observed;
   Vector   freqs;
   IntArray rareAlleles;

   for (int m = 0; m < Person::markerCount; m++)
      {
      int alleleCount = ped.CountAlleles(m);

      if (alleleCount <= 1) continue;

      observed.Dimension(alleleCount, alleleCount);
      observed.Zero();

      freqs.Dimension(alleleCount);
      freqs.Zero();

      rareAlleles.Dimension(alleleCount);
      rareAlleles.Zero();

      int total = 0;

      // First calculate allele frequencies
      for (int i = 0; i < ped.count; i++)
         if (ped[i].isGenotyped(m) &&
             (!ped.chromosomeX || ped[i].sex != SEX_MALE) &&
             (!foundersOnly || ped[i].isFounder()))
            {
            int one = ped[i].markers[m].one - 1;
            int two = ped[i].markers[m].two - 1;

            freqs[one]++;
            freqs[two]++;
            total++;
            }

      if (total <= 3.0) continue;

      // Work out allele frequencies
      freqs.Multiply(1.0 / (2.0 * total));

      // Find the minimum useful allele frequency
      double minFreq = sqrt(3.0 / total);

      // List all rare alleles and find their total frequency
      double rareSum = 0.0;

      // Also find the rarest allele with frequency >= minFreq
      int mini = -1, commonCount = 0;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] == 0.0)
            continue;
         else if (freqs[i] < minFreq)
            {
            rareSum += freqs[i];
            rareAlleles[i] = true;
            }
         else
            {
            if (mini == -1 || freqs[i] < freqs[mini])
               mini = i;
            commonCount++;
            }

      // If all rare alleles combined are still too rare...
      if (rareSum > 0.0 && rareSum < minFreq && commonCount > 1)
         {
         rareSum += freqs[mini];
         rareAlleles[mini] = true;
         }

      // Glob all rare alleles into 1
      int glob = alleleCount - 1;

      if (rareSum > 0.0)
         while (rareAlleles[glob] == false)
            glob--;

      // Recount allele and genotype frequencies globbing rare alleles as needed
      freqs.Zero();

      for (int i = 0; i < ped.count; i++)
         if (ped[i].isGenotyped(m) &&
             (!ped.chromosomeX || ped[i].sex != SEX_MALE) &&
             (!foundersOnly || ped[i].isFounder()))
            {
            int one = ped[i].markers[m].two - 1;
            int two = ped[i].markers[m].one - 1;

            if (rareAlleles[one]) one = glob;
            if (rareAlleles[two]) two = glob;

            if (one > two)
               observed[one][two]++;
            else
               observed[two][one]++;
            freqs[one]++;
            freqs[two]++;
            }

      freqs.Multiply(1.0 / (2.0 * total));

      double chisq = 0.0;
      bool   low_counts = false;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] > 0)
            for (int j = 0; j <= i; j++)
               if (freqs[j] > 0)
                  {
                  double expected = freqs[i] * freqs[j] * total * (i == j ? 1 : 2);
                  chisq += square(expected - observed[i][j]) / expected;
                  low_counts |= expected < 3.0;
                  }

      note |= low_counts;
      int realAlleles = 0;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] != 0.0)
            realAlleles++;

      double pvalue = realAlleles == 1.0 ? 1.0 :
                      chidist(chisq, realAlleles * (realAlleles - 1) * 0.5);

      if (pvalue > 0.05) continue;

      if (!fail)
         printf("%15s %7s %7s %7s %7s %7s\n",
                "", "N", "Alleles", "Largest", "ChiSq", "P-value");

      fail = true;

      printf("%15.15s %7d %7d %7d %7.2f %7.4f %s",
             (const char *) Person::markerNames[m],
             total, realAlleles, alleleCount, chisq, pvalue,
             low_counts ? "see note " : "");

      if (rareAlleles.Sum() > 1)
         printf("(%d alleles grouped)\n", rareAlleles.Sum());
      else
         printf("\n");
      }

   if (note && fail)
      printf("\nNOTE: Some genotypes had very low expecteds counts,\n"
             "      making this HW test unreliable.\n");

   printf(!fail ? "All markers okay.\n\n" : "\n");
   }

bool isParent(Person * a, Person * b)
   {
   Person * ancestor = (a->traverse < b->traverse) ? a : b;
   Person * descendant = (a->traverse < b->traverse) ? b : a;

   return isOrderedParent(ancestor, descendant);
   }

bool isOrderedParent(Person *a, Person *b)
   {
   return (!b->isFounder() && (b->father == a || b->mother == a));
   }

bool isGrandParentOffspring(Person * a, Person * b)
   {
   Person * ancestor = (a->traverse < b->traverse) ? a : b;
   Person * descendant = (a->traverse < b->traverse) ? b : a;

   return isOrderedGrandParentOffspring(ancestor, descendant);
   }

bool isOrderedGrandParentOffspring(Person *a, Person *b)
   {
   if (b->isFounder()) return false;

   if (!b->father->isFounder() && (b->father->father == a || b->father->mother == a))
       return true;

   if (!b->mother->isFounder() && (b->mother->father == a || b->mother->mother == a))
       return true;

   return false;
   }

bool isAvuncular(Person * a, Person * b)
   {
   return (isOrderedAvuncular(a,b) || isOrderedAvuncular(b, a));
   }

bool isOrderedAvuncular(Person *a, Person *b)
   {
   if (!b->isFounder() && !isParent(a, b))
      if (a->isSib(*b->father) || a->isSib(*b->mother))
         return true;

   return false;
   }

bool isSib(Person *a, Person *b)
   {
   if (a == b) return false;

   return a->isSib(*b);
   }

bool isHalfSib(Person *a, Person *b)
   {
   return a->isHalfSib(*b);
   }

bool isCousin(Person *a, Person *b)
   {
   if (a->isFounder() || b->isFounder() || a->isSib(*b) || a->isHalfSib(*b) )
      return false;

   if (!a->mother->isFounder())
      if (!b->mother->isFounder() && a->mother->isSib(*b->mother) ||
          !b->father->isFounder() && a->mother->isSib(*b->father) )
         return true;

   if (!a->father->isFounder())
      if (!b->mother->isFounder() && a->father->isSib(*b->mother) ||
          !b->father->isFounder() && a->father->isSib(*b->father) )
         return true;

   return false;
   }

void RelativePairs(bool (*filter)(Person &, Person &))
   {
   int parents = 0, sibs = 0, halfsibs = 0;
   int uncles = 0, cousins = 0, grandparents = 0;

   for (int f = 0; f < ped.familyCount; f++)
      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         for (int j = i + 1; j <= ped.families[f]->last; j++)
            if (filter == NULL || filter(ped[i], ped[j]))
               if (isParent(&ped[i], &ped[j])) parents++;
               else if (ped[i].isSib(ped[j])) sibs++;
               else if (ped[i].isHalfSib(ped[j])) halfsibs++;
               else if (isAvuncular(&ped[i], &ped[j])) uncles++;
               else if (isCousin(&ped[i], &ped[j])) cousins++;
               else if (isGrandParentOffspring(&ped[i], &ped[j])) grandparents++;

   if (filter != NULL)
      {
      printf("%12d %5d %9d %11d %10d %8d\n",
             parents, sibs, halfsibs, grandparents, uncles, cousins);
      return;
      }

   if (parents == 0)
      printf(" No relative pairs\n\n");
   else
      {
      printf(" Relative Pair Counts\n"
             "%25s: %d pairs\n", "Parent-Child", parents);
      if (sibs) printf("%25s: %d pairs\n", "Sib-pairs", sibs);
      if (halfsibs) printf("%25s: %d pairs\n", "Half-Sibs", halfsibs);
      if (grandparents) printf("%25s: %d pairs\n", "Grandparent-Grandchild", grandparents);
      if (uncles) printf("%25s: %d pairs\n", "Avuncular", uncles);
      if (cousins) printf("%25s: %d pairs\n", "Cousins", cousins);
      printf("\n");
      }
   }



 
 
