////////////////////////////////////////////////////////////////////// 
// merlin/MerlinCore.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday June 11, 2003
// 
 
#include "MerlinCore.h"
#include "Error.h"

#include <math.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>

// Limits to stop Merlin from attempting problems that are too tough
int MerlinCore::maxBits = 24;
int MerlinCore::maxMinutes = 0;

// Positions to analyse
int    MerlinCore::steps_per_interval = 0;
double MerlinCore::maxDistance = _NAN_;
double MerlinCore::minDistance = _NAN_;
double MerlinCore::grid  = _NAN_;
double MerlinCore::start = _NAN_;
double MerlinCore::stop  = _NAN_;

// Output options
bool MerlinCore::quietOutput = false;
bool MerlinCore::useMarkerNames = false;

// Swap options
bool MerlinCore::useSwap = false;

// Internal flags to minimize useless calculations
bool MerlinCore::multipoint = false;
bool MerlinCore::scanChromosome = false;
bool MerlinCore::betweenMarkers = false;

// Flags for controlling recombination
bool MerlinCore::zeroRecombination = false;
bool MerlinCore::oneRecombination = false;
bool MerlinCore::twoRecombinations = false;
bool MerlinCore::threeRecombinations = false;
bool MerlinCore::twopoint = false;

// Determines whether error checking stub should be executed
bool MerlinCore::findErrors = false;

// Constructor and Destructor
//

MerlinCore::MerlinCore(Pedigree & pedigree) : mantra(), ped(pedigree)
   {
   printHeader = false;
   lowBound = pow(2.0, -129);
   rescale  = pow(2.0, 258);
   lnScale  = 258 * log(0.5);
   }

MerlinCore::~MerlinCore()
   {
   }

// Global setup that must be carried out before calculating likelihods
//

void MerlinCore::SetupGlobals()
   {
   // Set the maximum number of recombinants in multipoint calculations
   Multipoint::maximum_recombinants =
      threeRecombinations * 3 + twoRecombinations * 2 + oneRecombination;
   }

// Setup map for analysing the next chromosome at currently selected locations
//

int MerlinCore::SetupMap(int chromosome)
   {
   markers.Dimension(0);
   int next = PedigreeGlobals::SortMarkersInMapOrder(markers, chromosome);

   markerCount = markers.Length();

   if (markerCount == 0) error("No markers could be placed");

   // Clear previous information
   labels.Clear();
   analysisPositions.Clear();
   markerPositions.Clear();

   String label;

   // Retrieve marker positions
   for (int i = 0; i < markerCount; i++)
      {
      MarkerInfo * info = PedigreeGlobals::GetMarkerInfo(markers[i]);
      ValidateMarker(info);
      markerPositions.Push(info->position);
      }

   // For maps with no recombination, only analyse one location
   if (zeroRecombination)
      {
      analysisPositions.Push(markerPositions[0]);
      labels.Push(useMarkerNames ?
                  PedigreeGlobals::GetMarkerInfo(markers[0])->name :
                  label = (markerPositions[0] * 100.0));
      }
   // Singlepoint analysis at marker locations only
   else if (twopoint)
      {
      analysisPositions = markerPositions;

      for (int i = 0; i < markerCount; i++)
         labels.Push(PedigreeGlobals::GetMarkerInfo(markers[i])->name);
      }
   // Analysis along grid of equally spaced locations
   else if (grid != _NAN_)
      {
      double position = start != _NAN_ ? start * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers[0])->position;
      double mapEnd = stop != _NAN_ ? stop * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers.Last())->position;
      double step = grid * 0.01;

      while (true)
         {
         analysisPositions.Push(position);
         labels.Push(label = position * 100.0);
         if (mapEnd - position < 1e-5) break;
         position += step;
         }
      }
   // Analysis at marker positions and equally spaced locations in-between
   else
      {
      double min = minDistance == _NAN_ ? _NAN_ : minDistance * 0.01;
      double max = maxDistance == _NAN_ ? _NAN_ : maxDistance * 0.01;

      double last = markerPositions[0];

      for (int i = 0; i < markerCount; i++)
         {
         if (markerPositions[i] > last)
            {
            double delta = markerPositions[i] - last;
            int    steps = steps_per_interval;
            double step  = delta / (steps + 1);

            if ((max != _NAN_) && (max < step))
               {
               steps = (int) (delta / (max + 1e-6));
               step  = delta / (steps + 1);
               }

            if ((min != _NAN_) && min > step)
               steps = (int) (delta / (minDistance  + 1e-6) - 0.999);

            if (steps < 0) steps = 0;

            for (int j = 1; j <= steps; j++)
               {
               double position = last + delta * j / (steps+1);

               analysisPositions.Push(position);
               labels.Push(label = position * 100.0);
               }
            }

         if (markerPositions[i] > last || i == 0)
            {
            analysisPositions.Push(last = markerPositions[i]);
            labels.Push(useMarkerNames ?
               PedigreeGlobals::GetMarkerInfo(markers[i])->name :
               label = (last * 100.0));
            }
         }
      }
   return next;
   }

// Setup analyses for a new family
//

bool MerlinCore::SelectFamily(Family * f)
   {
   mantra.Prepare(ped, *f);
   mantra.PrepareIBD();

   printHeader = true;
   cleanOutput = false;

   family = f;
   serial = f->serial;

   // if (mantra.couples)
   //   PrintMessage("  %d symmetric founder couple%s detected",
   //               mantra.couples, mantra.couples == 1 ? "" : "s");

   if (mantra.bit_count <= maxBits)
      return true;

   PrintMessage("  SKIPPED: Maximum complexity set at %d bits", maxBits);
   CleanMessages();
   return false;
   }

// Calculation of single marker likelihoods
//

void MerlinCore::ScoreSinglepoint()
   {
   singlepoint = new InheritanceTree[ped.markerCount];
   InheritanceTree tempVectors;

   informativeCount = 0;
   informativeMarkers.Clear();
   informativePositions.Clear();
   informationScores.Clear();
   informationBounds.Clear();

#ifdef __PRINTOUT_STATS__
   Vector nodeCounts, leafCounts, zeroCounts, infoScores;
#endif

   // Reset the likelihood
   likelihood = 0;

   for (int m = 0; m < markerCount; m++)
      {
      ProgressReport("Singlepoint", m, markerCount);
      tempVectors.ScoreVectors(mantra, markers[m]);

#ifndef __PRINTOUT_STATS__
      stats.GetTreeInfo(tempVectors);
#else
      stats.GetDetailedInfo(tempVectors);
      nodeCounts.Push(stats.nodeCount);
      leafCounts.Push(stats.leafCount);
      zeroCounts.Push(stats.zeroCount);
      infoScores.Push(stats.information);
#endif

      if (stats.mean == 0.0)
         {
         // TODO -- be helpful and try to pin down problem genotypes
         PrintMessage("  Skipping Marker %-10s [BAD INHERITANCE]",
              (const char *) ped.markerNames[markers[m]]);

         // Replace with uninformative inheritance tree
         tempVectors.MakeMinimalTree(1.0, mantra.bit_count);
         stats.information = 0.0;

         // Erase genotypes for this marker
         for (int i = family->first; i <= family->last; i++)
            ped[i].markers[markers[m]][2] = ped[i].markers[markers[m]][1] = 0;
         }

      if (stats.information > 1e-5 || twopoint ||
           InheritanceTree::aggressiveMerging == false &&
           tempVectors.nodes[0].type != TREE_NODE_ZERO &&
           tempVectors.nodes[0].type != TREE_NODE_LEAF)
         {
         singlepoint[informativeCount].Copy(tempVectors);
         singlepoint[informativeCount++].Pack();
         informativeMarkers.Push(markers[m]);
         informativePositions.Push(markerPositions[m]);
         informationScores.Push(stats.information);
         informationBounds.Push(stats.information > 0.3 ? stats.GetMinInfo(tempVectors) : 0.0);
         }
      else
         {
         // Likelihood for all gene flow patterns is effectively equal
         if (stats.information > 1e-10 && !quietOutput)
            PrintMessage("  Skipping Marker %-10s [Info = %.6f]",
                (const char *) ped.markerNames[markers[m]], stats.information);

         // stats.mean as calculated by GetTreeInfo must be scaled
         if (stats.mean)
            likelihood += log(stats.mean) + mantra.bit_count * log(0.5);
         }
      }

   if (Multipoint::maximum_recombinants || twopoint || zeroRecombination ||
       mantra.bit_count < 3)
      useSparse = false;
   else
      {
      // By default we don't assume sparseness
      useSparse = false;

      // Lower information bounds if founder couple symmetries are being used
      if (mantra.couples)
         // TODO -- work out what a better penalty for couples is ...
         informationBounds.Add(-1.0 * mantra.couples / mantra.bit_count);

      // Unless very informative marker pairs are present
      for (int i = 0; i < informativeCount - 1; i++)
         if (informationBounds[i] + informationBounds[i + 1] > 1.0)
            useSparse = true;
      }

#ifdef __PRINTOUT_STATS__
   int median = nodeCounts.dim / 2;
   int ciLo = nodeCounts.dim * 25 / 1000;
   int ciHi = nodeCounts.dim * 975 / 1000;

   char * labels [] = { "Nodes", "Leafs", "Zeros", "Info" };
   Vector * vect [] = { &nodeCounts, &leafCounts, &zeroCounts, &infoScores};

   for (int i = 0; i < 4; i++)
      {
      Vector & vector = *vect[i];

      vector.Sort();
      PrintMessage("  % 8s: %.2f (%.2f to %.2f) Median = %.2f, 95%% CI = %.2f - %.2f",
         labels[i], vector.Average(), vector.Min(), vector.Max(),
         vector[median], vector[ciLo], vector[ciHi]);
      }
#endif
   }

// Calculation of right conditional likelihoods
//

bool MerlinCore::ScoreConditionals()
   {
   if (zeroRecombination)
      return ZeroRecombination();

   if (twopoint) return true;

   right = new Tree[informativeCount];
   rightScale.Dimension(informativeCount);
   leftScale.Dimension(informativeCount);

   Multipoint engine;

   if (useSparse) engine.SetupConditioning(mantra);

   int last = informativeCount - 1;

   rightScale[last] = 1.0;
   singlepoint[last].UnPack();
   right[last].Copy(singlepoint[last]);
   singlepoint[last].Discard();

   bool   check_time = (mantra.bit_count >= 16) && (maxMinutes > 0);
   double seconds;
   time_t start_time, curr_time;

   if (check_time)
      {
      seconds = maxMinutes * 60.;
      // It would be good to check for problems in time call, but
      // errno is not actually part of ANSI specification
      // errno = 0;
      time(&start_time);
      // if (errno) check_time = false;
      }

   int rescalings = 0;

   for (int m = last - 1; m >= 0; m--)
      {
      ProgressReport("Right Conditional", last - m, informativeCount);

      // Calculate recombination fraction
      double distance = informativePositions[m+1]-informativePositions[m];

      // Retrieve inheritance at m+1 given m + 1, m + 2, ...
      engine.Clear();
      engine.Copy(right[m+1]);
      right[m+1].Pack();

      if (useSparse && (informationBounds[m+1] + informationBounds[m] > 1.0))
         {
         // Try fast conditioning for informative locations?
         singlepoint[m].UnPack();
         right[m].Copy(singlepoint[m]);
         singlepoint[m].Discard();

         // Calculate conditional probability without intermediates
         engine.Condition(right[m], distance, rightScale[m+1]);
         }
      else
         {
         // Calculate inheritance at m given m+1, m+2, ...
         engine.MoveAlong(mantra, distance, rightScale[m+1]);

         // Calculate inheritance at m given *** m ***, m + 1, ...
         singlepoint[m].UnPack();
         engine.Multiply(singlepoint[m]);
         singlepoint[m].Discard();

         // Store it for future use
         right[m].Copy(engine);
         }

      // Check that the likelihod hasn't hit zero
      if (stats.GetMean(right[m]) == 0.0) return false;

      // If the tree is very unlikely, multiply by a constant to avoid underflow
      rightScale[m] = stats.mean < lowBound ? (rescalings+=m!=0, rescale) : 1.0;

      if (check_time)
         if (difftime(time(NULL), start_time) > seconds)
            throw OutOfTime();
      }

   if (informativeCount == 1) stats.GetMean(right[0]);
   likelihood += log(stats.mean) + rescalings * lnScale;

   right[0].Pack();
   return true;
   }

// Likelihood calculation assuming no recombination
//

void MerlinCore::SortInformation(int start_index, int end_index)
   {
   if (start_index >= end_index) return;

   double average = (informationScores[informationIndex[start_index]] +
                     informationScores[informationIndex[end_index]]) * 0.5;

   int a = start_index, b = end_index;

   while (true)
      {
      while ((informationScores[informationIndex[a]] >= average) && a < b)
         a++;

      while ((informationScores[informationIndex[b]] <= average) && a < b)
         b--;

      if (a == b) break;

      informationIndex.Swap(a, b);
      }

   SortInformation(start_index, a == end_index ? a - 1 : a);
   SortInformation(a == start_index ? a + 1 : a, end_index);
   }

bool MerlinCore::ZeroRecombination()
   {
   // Identify the most informative markers
   informationIndex.Dimension(informativeCount);
   informationIndex.SetSequence(0, 1);
   SortInformation(0, informativeCount - 1);

   // Combined tree includes all nodes that are non-zero in all trees
   Tree combined;

   // Allocate storage for final result
   right = new Tree[1];
   singlepoint[informationIndex[0]].UnPack();
   right[0].Copy(singlepoint[informationIndex[0]]);
   singlepoint[informationIndex[0]].Discard();

   // Keep track of how many likelihood rescalings were carried out
   int rescalings = 0;

   // Locate inheritance vectors that are compatible with all markers
   for (int i = 1; i < informativeCount; i++)
      {
      // Incorporate next most informative marker
      combined.Copy(right[0]);
      singlepoint[informationIndex[i]].UnPack();
      combined.Multiply(singlepoint[informationIndex[i]]);
      singlepoint[informationIndex[i]].Discard();
      combined.TrimNoMerge();
      right[0].Copy(combined);

      // Check that is possible given zero recombination
      if (stats.GetMean(right[0]) == 0.0)
         return false;

      // Avoid underflow by multiplying by large rescale constant
      if (stats.mean < lowBound)
         {
         rescalings++;
         right[0].Multiply(0, rescale);
         }
      }

   likelihood += log(stats.mean) + rescalings * lnScale;

   right[0].Pack();
   return true;
   }

// Calculation of right conditional likelihoods
//

void MerlinCore::WalkLeft
   (Multipoint & left, Multipoint & leftMarker, Multipoint & full, int marker)
   {
   // Load required information to memory
   singlepoint[marker+1].UnPack();
   right[marker+1].UnPack();

   // Calculate the new left conditioned probabilities
   if (marker != -1)
      {
      // First, check how far from the previous location
      double distance = informativePositions[marker + 1] -
                        informativePositions[marker];

      // Load previous left conditioned likelihoods
      left.Copy(leftMarker);

      // Load the singlepoint likelihoods at the next location
      leftMarker.Copy(singlepoint[marker + 1]);

      if (!findErrors && useSparse &&
         (informationBounds[marker] + informationBounds[marker + 1] > 1.0))
         {
         // Build left conditioned probability up to marker + 1 (inclusive)
         left.Condition(leftMarker, distance, leftScale[marker]);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // TODO -- We could choose to condition on left / right ?
         // Get right conditioned probability
         full.Copy(right[marker + 1]);
         left.Condition(full, distance, leftScale[marker]);
         }
      else
         {
          // Do the multipoint calculation
         left.MoveAlong(mantra, distance, leftScale[marker]);

         // Get the full left conditioned probability
         leftMarker.Multiply(left);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // Get the full left and right conditioned probability
         full.Copy(right[marker + 1]);
         full.Multiply(left);
         }
       }
   else
      {
      // This is for the first marker
      leftMarker.Copy(singlepoint[0]);
      full.Copy(right[0]);
      leftScale[0] = 1.0;
      }

   // Check the marker for errors
   if (findErrors) ErrorCheck(left, full, marker + 1);

   // Free up memory
   singlepoint[marker+1].Discard();
   right[marker+1].Discard();
   }

// Scan chromosome and calculate likelihoods at each location for analyses
//

void MerlinCore::ScanChromosome()
   {
   // Anchors for starting position
   int leftAnchor  = -1;
   int rightAnchor =  0;
   int positionCount = analysisPositions.dim;

   // Left-conditioned probabilities up to left-marker
   Multipoint leftMarker;

   // Left and right conditionals up to current position
   Multipoint leftInheritance, inheritance;

   // Setup auxiliary variables for processing very sparse trees
   if (useSparse) leftInheritance.SetupConditioning(mantra);

   for (int i = 0; i < positionCount; i++)
      {
      ProgressReport("Scanning Chromosome", i, positionCount);

      double position = analysisPositions[i];

      // Move anchors so they flank the current position
      while (!zeroRecombination && !twopoint &&
             rightAnchor < informativeCount &&
             informativePositions[rightAnchor] <= position)
         {
         WalkLeft(leftInheritance, leftMarker, inheritance, leftAnchor++);
         rightAnchor++;
         }

      // Special case of two point analysis
      if (twopoint)
         {
         rightAnchor = leftAnchor = i;
         singlepoint[i].UnPack();
         inheritance.Copy(singlepoint[i]);
         singlepoint[i].Discard();
         }
      // If we are analysing at marker, the hard work is done
      else if (leftAnchor != -1 && informativePositions[leftAnchor] == position)
          {
          // nothing to do!
          }
      else if (!betweenMarkers)
         continue;
      // If there are no left flanking informative markers
      else if (leftAnchor == -1)
         {
         double distance = informativePositions[rightAnchor] - position;

         right[rightAnchor].UnPack();
         inheritance.Copy(right[rightAnchor]);
         inheritance.MoveAlong(mantra, distance);
         right[rightAnchor].Discard();
         }
      // If there are no right flanking informative markers
      else if (rightAnchor == informativeCount)
         {
         double distance = position - informativePositions[leftAnchor];

         inheritance.Copy(leftMarker);
         inheritance.MoveAlong(mantra, distance);
         }
      // Otherwise we are really between markers
      else
         {
         double distance = position - informativePositions[leftAnchor];

         leftInheritance.Copy(leftMarker);
         leftInheritance.MoveAlong(mantra, distance, leftScale[leftAnchor]);

         distance = informativePositions[rightAnchor] - position;

         right[rightAnchor].UnPack();
         inheritance.Copy(right[rightAnchor]);
         inheritance.MoveAlong(mantra, distance, rightScale[rightAnchor]);
         inheritance.Multiply(leftInheritance);
         right[rightAnchor].Discard();
         }

      AnalyseLocation(i, inheritance);
      }
   }

// Output buffering functions
//

void MerlinCore::PrintHeader()
   {
   printHeader = false;

   printf("Family: %5s - Founders: %-2d - Descendants: %-2d - Bits: %-2d\n",
          (const char *) family->famid, mantra.f, mantra.n - mantra.f,
          mantra.bit_count);
   }

void MerlinCore::PrintMessage(const char * msg, ...)
   {
   if (printHeader) PrintHeader();

   va_list ap;

   va_start (ap, msg);
   vprintf(msg, ap);
   printf("\n");
   va_end(ap);
   }

void MerlinCore::ProgressReport(const char * msg, int done, int total)
   {
   if (quietOutput) return;

   long check = 0x10000 >> mantra.bit_count;
   if (mantra.bit_count > 16 || done & check )
      {
      if (printHeader)
         printf("Family %s (%d bits) -",
               (const char *) family->famid, mantra.bit_count);
      printf("  %s: %3d%%%20s\r", msg, done * 100 / total, "");
      fflush(stdout);
      cleanOutput = true;
      }
   }

void MerlinCore::CleanMessages()
   {
   if (cleanOutput)
      {
      printf("%70s\r", "");
      fflush(stdout);
      }

   if (!printHeader)
      printf("\n");
   }

// Enforce maximum number of alleles on markers
//

void MerlinCore::ValidateMarker(MarkerInfo * info)
   {
#ifdef __USE_LONG_INT
   if (info->freq.dim > 64)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin supports up to 64 alleles",
             (const char *) info->name, info->freq.dim - 1);
#else
   if (info->freq.dim > 32)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin supports up to 32 alleles",
             (const char *) info->name, info->freq.dim - 1);
#endif
   }

// Stub functions that should be replaced by derived classes

void MerlinCore::AnalyseLocation(int, Multipoint &) { }
void MerlinCore::ErrorCheck(Multipoint &, Multipoint &, int ) { }


 
 
