////////////////////////////////////////////////////////////////////// 
// merlin/MerlinFamily.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday June 11, 2003
// 
 
#include "MerlinFamily.h"
#include "MerlinHaplotype.h"
#include "MapFunction.h"
#include "MathConstant.h"
#include "NPL-ASP.h"
#include "NPL-QTL.h"

// General analyses
bool FamilyAnalysis::estimateIBD = false;
bool FamilyAnalysis::estimateKinship = false;
bool FamilyAnalysis::estimateKinship15 = false;
bool FamilyAnalysis::estimateInformation = false;
bool FamilyAnalysis::estimateMatrices = false;
bool FamilyAnalysis::storeKinship = false;
bool FamilyAnalysis::calcLikelihood = false;

// Haplotyping
bool FamilyAnalysis::bestHaplotype = false;
int  FamilyAnalysis::sampledHaplotypes = 0;
bool FamilyAnalysis::allHaplotypes = false;

// Non-parametric linkage analysis
bool FamilyAnalysis::nplAll = false;
bool FamilyAnalysis::nplPairs = false;
bool FamilyAnalysis::nplQtl = false;
bool FamilyAnalysis::nplDeviates = false;

// Output options
bool FamilyAnalysis::simwalk2 = false;
bool FamilyAnalysis::perFamily = false;
bool FamilyAnalysis::writePDF = false;

FamilyAnalysis::FamilyAnalysis(Pedigree & pedigree) :
   MerlinCore(pedigree), kinship(mantra), kinship15(mantra),
   ibd(mantra), matrix(mantra), kac(pedigree, labels), hybrid(*this),
   pdf(analysisPositions)
   {
   nplCount = 0;
   npl = NULL;
   errorfile = infofile = NULL;
   }

FamilyAnalysis::~FamilyAnalysis()
   {
   if (npl != NULL) delete [] npl;
   }

bool FamilyAnalysis::SelectFamily(Family * f)
   {
   if (MerlinCore::SelectFamily(f))
      // If analysis for this family is proceeding ...
      {
      if (estimateMatrices) matrix.SelectFamily(&ped, family);
      if (estimateIBD) ibd.SelectFamily(&ped, family);
      if (estimateKinship || storeKinship) kinship.SelectFamily(&ped, family);
      if (estimateKinship15) kinship15.SelectFamily(&ped, family);

      return true;
      }

   // Tell simwalk2 we have skipped this family
   if (simwalk2) hybrid.Skip();

   return false;
   }

void FamilyAnalysis::SetupGlobals()
   {
   MerlinCore::SetupGlobals();

   // How many NPL statistics to calculate
   nplCount = (nplAll + nplPairs) * ped.affectionCount +
              (nplQtl + nplDeviates) * ped.traitCount;

   if (npl != NULL) delete [] npl;

   nplInv.Dimension(nplCount);
   nplMean.Dimension(nplCount);
   npl = new Tree[nplCount];

   // Carry out between marker analyses?
   betweenMarkers = estimateIBD || estimateKinship || estimateKinship15 ||
      estimateInformation || estimateMatrices || storeKinship || nplCount != 0;

   // Carry out analyses at each marker?
   scanChromosome = betweenMarkers || findErrors;

   // Perform right conditional calculations?
   multipoint     = scanChromosome || bestHaplotype && zeroRecombination ||
                    (sampledHaplotypes != 0) || allHaplotypes || calcLikelihood;

   // Merge gene flow outcomes with identical likelihoods, but possibly
   // implying different allelic states
   InheritanceTree::aggressiveMerging = !allHaplotypes;

   // Set up labels for NPL output
   kac.LabelPhenotypes(nplAll, nplPairs, nplQtl, nplDeviates);

   // Check that we have the correct type of data for simwalk2 analyses
   if (simwalk2 && ped.affectionCount != 1)
      error("Joint analysis with SimWalk2 requires pedigree with exactly one discrete trait\n"
            "The current pedigree includes %d discrete traits\n", ped.traitCount);
   }

int FamilyAnalysis::SetupMap(int chromosome)
   {
   int next_chromosome = MerlinCore::SetupMap(chromosome);

   // Setup chromosome labels for PDF output
   if (writePDF)
#ifdef __CHROMOSOME_X__
      pdf.SelectChromosomeX();
#else
      pdf.SelectChromosome(ped.GetMarkerInfo(markers[0])->chromosome);
#endif

   // Allocate array for storing information content statistics
   information.Dimension(2, analysisPositions.dim);
   information.Zero();

   // Allocates array for storing NPL scores for each location in each family
   if (nplCount != 0) kac.Prepare(nplCount, ped.familyCount);

   // Discards any previous kinship coefficients
   if (storeKinship) kinship.SetFamilyCount(ped.familyCount);

   // Initialize overall likelihood for this chromosome
   lkSum = 0.0;
   lkCount = 0;

   return next_chromosome;
   }

void FamilyAnalysis::Analyse()
   {
   singlepoint = NULL;
   right = NULL;

   try
      {
      BasicTree::FreeSwap();

      ScoreSinglepoint();

      if (informativeCount == 0)
         SkipFamily();
      else
         {
         if (storeKinship) kinship.SetPositionCount(analysisPositions.dim);
         if (nplCount) ScoreNPL();
         if (multipoint)
            {
            if (ScoreConditionals())
               {
               if (scanChromosome) ScanChromosome();
               if (calcLikelihood && perFamily)
                  PrintMessage("  lnLikelihood = %.3f", likelihood);
               if (allHaplotypes) haplo.All(*this);
               for (int i = 0; i < sampledHaplotypes; i++) haplo.Sample(*this);
               if (bestHaplotype && zeroRecombination) haplo.MostLikely(*this);
               lkSum += likelihood;
               lkCount ++;
               }
            else
               {
               PrintMessage("  SKIPPED: Requires impossible recombination pattern");

               if (nplCount != 0) kac.SkipFamily(family->serial);
               if (storeKinship) kinship.Discard();
               if (simwalk2) nplInv[0] = nplInv[1] = 0.0;
               }
            if (!twopoint) delete [] right;
            if (simwalk2) hybrid.Output();
            }
         if (bestHaplotype && !zeroRecombination) haplo.MostLikely(*this);
         }

      CleanMessages();
      delete [] singlepoint;
      }
   catch (const TreesTooBig & problem)
      {
      PrintMessage("  SKIPPED: At least %d megabytes needed", problem.memory_request);
      CleanMessages();

      if (singlepoint != NULL) delete [] singlepoint;
      if (right != NULL) delete [] right;

      if (simwalk2) hybrid.Skip();
      if (nplCount != 0) kac.SkipFamily(family->serial);
      if (storeKinship) kinship.Discard();
      }
   catch (const OutOfTime & problem)
      {
      PrintMessage("  SKIPPED: Analysis would require more than %d minutes\n", maxMinutes);
      CleanMessages();

      if (singlepoint != NULL) delete [] singlepoint;
      if (right != NULL) delete [] right;

      if (simwalk2) hybrid.Skip();
      if (nplCount != 0) kac.SkipFamily(family->serial);
      if (storeKinship) kinship.Discard();
      };
   };

void FamilyAnalysis::SkipFamily()
   {
   // Print out overall gene-flow likelihood
   if (calcLikelihood && perFamily)
      PrintMessage("  lnLikelihood = %.3f", likelihood);

   // And update overall likelihood for the sample
   lkSum += likelihood;
   lkCount++;

   // If there is no information on this family
   // IBD and kinship coefficients / matrices are the same at all markers
   if (estimateIBD || estimateKinship || estimateKinship15 || estimateMatrices || storeKinship)
      {
      double sum;
      Tree dummy;

      dummy.MakeMinimalTree(1.0, mantra.bit_count);

      double scale = pow(2.0, mantra.bit_count);

      if (storeKinship) kinship.SetPositionCount(labels.Length());
      if (estimateIBD) sum = ibd.Calculate(dummy, labels[0]) * scale;
      if (estimateMatrices ) sum = matrix.Calculate(dummy, labels[0]) * scale;
      if (estimateKinship || storeKinship) sum = kinship.Calculate(dummy) * scale;
      if (estimateKinship) kinship.Output(labels[0], sum);
      if (estimateKinship15) sum = kinship15.Calculate(dummy, labels[0]) * scale;
      if (storeKinship) kinship.Store(0, sum);

      for (int i = 1; i < labels.Length(); i++)
         {
         if (estimateIBD) ibd.Output(labels[i], sum);
         if (estimateKinship) kinship.Output(labels[i], sum);
         if (estimateKinship15) kinship15.Output(labels[i], sum);
         if (estimateMatrices) matrix.Output(labels[i], sum);
         if (storeKinship) kinship.Store(i, sum);
         }
      }

   // Output dummy haplotype -- something to look at -- keeps people happy
   if (allHaplotypes || bestHaplotype) haplo.UninformativeFamily(*this);

   for (int i = 0; i < sampledHaplotypes; i++)
      haplo.UninformativeFamily(*this);

   // And tell Simwalk2 this family isn't very interesting (to us!)
   if (simwalk2) hybrid.Skip();
   }

void FamilyAnalysis::ScoreNPL()
   {
   int         index = 0;

   if (nplAll)
      {
      NPL_ALL engine;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-ALL statistic", a, ped.affectionCount);

         engine.ScoreNPL(mantra, a);
         npl[index++].Copy(engine);
         }
      }

   if (nplPairs)
      {
      NPL_Pairs engine;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-Pairs statistic", a, ped.affectionCount);

         engine.ScoreNPL(mantra, a);
         npl[index++].Copy(engine);
         }
      }

   if (nplQtl)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-QTL statistic", t, ped.traitCount);

         engine.ScoreWithSampleMean(mantra, t);
         npl[index++].Copy(engine);
         }
      }

   if (nplDeviates)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-Deviate statistic", t, ped.traitCount);

         engine.ScoreWithMeanZero(mantra, t);
         npl[index++].Copy(engine);
         }
      }

   while (index--)
      {
      stats.GetBounds(npl[index]);

      nplMean[index] = stats.mean;
      if (stats.var > TINY)
         {
         nplInv[index] = 1.0 / sqrt(stats.var);
         kac.min[index][serial] = (stats.min - stats.mean) * nplInv[index];
         kac.max[index][serial] = (stats.max - stats.mean) * nplInv[index];
         }
      else
         nplInv[index] = 0.0;
      }
   }

void FamilyAnalysis::AnalyseLocation(int pos, Multipoint & inheritance)
   {
   // The probability of the data summed all possible inheritance vectors
   double lk;

   if (estimateKinship15) kinship15.Calculate(inheritance, labels[pos]);
   if (estimateMatrices) lk = matrix.Calculate(inheritance, labels[pos]);
   if (estimateIBD) lk = ibd.Calculate(inheritance, labels[pos]);
   if (estimateKinship || storeKinship)
      lk = kinship.Calculate(inheritance,
                             estimateKinship ? (const char *) labels[pos] : NULL,
                             storeKinship ? pos : -1);
   if (!estimateIBD && !estimateKinship && !estimateMatrices && !storeKinship)
      lk = stats.GetMean(inheritance);

   for (int a = 0; a < nplCount; a++)
      // Only analyse informative families (eg. multiple related affecteds)
      if (nplInv[a] != 0.0)
         kac.scores[a][serial][pos] =
            (inheritance.MeanProduct(npl[a])/lk - nplMean[a]) * nplInv[a];

   if (estimateInformation)
      {
      // Calculate information
      stats.GetTreeInfo(inheritance);

      information[0][pos] += stats.information * mantra.bit_count;
      information[1][pos] += mantra.bit_count;

      if (infofile != NULL)
         fprintf(infofile, "%10s %10s %12.5f %5d\n",
            (const char *) family->famid, (const char *) labels[pos],
            stats.information, mantra.bit_count);
      }
   }

void FamilyAnalysis::ErrorCheck
     (Multipoint & left, Multipoint & full, int informativeMarker)
   {
   if (informativeCount <= 1) return;

   Multipoint      rightInheritance;
   InheritanceTree alternative;

   double singlepoint_likelihood= stats.GetMean(singlepoint[informativeMarker]);
   double multipoint_likelihood = stats.GetMean(full);

   int marker = informativeMarkers[informativeMarker];

   if (++informativeMarker != informativeCount)
      {
      double distance = informativePositions[informativeMarker] -
                        informativePositions[informativeMarker-1];
      right[informativeMarker].UnPack();
      rightInheritance.Copy(right[informativeMarker]);
      right[informativeMarker].Discard();
      rightInheritance.MoveAlong(mantra, distance, rightScale[informativeMarker]);
      if (informativeMarker != 1)
         rightInheritance.Multiply(left);
      }

   for (int error = family->first; error <= family->last; error++)
      if (ped[error].markers[marker].isKnown())
         {
         Alleles saved_genotype = ped[error].markers[marker];
         ped[error].markers[marker][0] = ped[error].markers[marker][1] = 0;

         alternative.ScoreVectors(mantra, marker);

         double alternative_likelihood = stats.GetMean(alternative);

         if (alternative_likelihood != singlepoint_likelihood)
            {
            double alternative_multipoint;

            if (informativeMarker == informativeCount)
               alternative_multipoint = alternative.MeanProduct(left);
            else
               alternative_multipoint = alternative.MeanProduct(rightInheritance);

            double score =
               (multipoint_likelihood / alternative_multipoint) /
               (singlepoint_likelihood / alternative_likelihood);

            if (score < 0.025)
               {
               PrintMessage("  %s genotype for individual %s is unlikely [%.3e]",
                           (const char *) ped.markerNames[marker],
                           (const char *) ped[error].pid, score);
               fprintf(errorfile, "%10s %10s %10s %#10.3g\n",
                       (const char *) family->famid,
                       (const char *) ped[error].pid,
                       (const char *) ped.markerNames[marker], score);
               }
            }

         ped[error].markers[marker] = saved_genotype;
         }
   }

void FamilyAnalysis::ShowInformation()
   {
   if (!estimateInformation) return;

   printf("%15s %6s\n", "Position", "Info");

   for (int i = 0; i < labels.Length(); i++)
      printf("%15s %6.4f\n", (const char *) labels[i],
             information[1][i] == 0.0 ? 0.0 :
             information[0][i] / information[1][i]);

   if (writePDF)
      {
      pdf.PrepareChart();

      pdf.chart.title = "Information Content";
      pdf.chart.yAxis.SetMin(0.0);
      pdf.chart.yAxis.SetMax(1.0);
      pdf.chart.yAxis.label = "Information Content";

      for (int i = 0; i < analysisPositions.Length(); i++)
         pdf.y[0][i] = information[1][i] == 0.0 ? 0.0 :
                                   information[0][i] / information[1][i];

      pdf.DrawChart();
      }

   printf("\n");
   }

void FamilyAnalysis::ShowLODs()
   {
   if (nplCount) kac.Analyse(pdf);
   if (storeKinship) vc.Analyse(*this);
   }

void FamilyAnalysis::OpenErrorFile()
   {
   errorfile = fopen("merlin.err", "wt");
   if (errorfile == NULL)
      error("Can't open error file [merlin.err]\n");
   fprintf(errorfile, "%10s %10s %10s %10s\n",
      "FAMILY", "PERSON", "MARKER", "RATIO");
   }

void FamilyAnalysis::OpenPerFamilyFiles()
   {
   if (estimateInformation)
      {
      infofile = fopen("merlin.inf", "wt");
      if (infofile == NULL)
         error("Can't open information file [merlin.inf]\n");
      fprintf(infofile, "%10s %10s %12s %5s\n", "FAMILY", "POSITION", "INFORMATION", "BITS");
      }
   if (nplCount) kac.OpenPerFamilyFiles();
   }

void FamilyAnalysis::SetupFiles()
   {
   if (findErrors) OpenErrorFile();
   if (perFamily) OpenPerFamilyFiles();
   if (estimateMatrices) matrix.OpenFile();
   if (estimateKinship) kinship.OpenFile();
   if (estimateKinship15) kinship15.OpenFile();
   if (estimateIBD) ibd.OpenFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.OpenFile();
   if (simwalk2) hybrid.OpenFiles();
   if (useSwap) BasicTree::SetupSwap();
   if (writePDF) pdf.OpenFile("merlin.pdf");
   }

void FamilyAnalysis::CloseFiles()
   {
   if (findErrors)
      {
      printf("Unlikely genotypes listed in file [merlin.err]\n");
      fclose(errorfile);
      }
   if (infofile != NULL)
      {
      printf("Information scores for individual families in file [merlin.inf]\n");
      fclose(infofile);
      }
   if (perFamily && nplCount) kac.ClosePerFamilyFiles();
   if (estimateMatrices) matrix.CloseFile();
   if (estimateKinship) kinship.CloseFile();
   if (estimateKinship15) kinship15.CloseFile();
   if (estimateIBD) ibd.CloseFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.CloseFile();
   if (simwalk2) hybrid.CloseFiles();
   if (useSwap) BasicTree::CloseSwap();
   if (writePDF) pdf.CloseFile();
   }

void FamilyAnalysis::ShowLikelihood()
   {
   if (!calcLikelihood) return;

   printf("lnLikelihood for %d families (%d skipped) = %.3f\n\n",
          lkCount, ped.familyCount - lkCount, lkSum);
   }
 
 
 
