////////////////////////////////////////////////////////////////////// 
// merlin/MerlinParameters.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday June 11, 2003
// 
 
#include "MerlinParameters.h"
#include "MerlinFamily.h"
#include "Pedigree.h"
#include "Random.h"

int  MerlinParameters::maxMegabytes = 0;
bool MerlinParameters::trimPedigree = false;
bool MerlinParameters::simulateNull = false;
bool MerlinParameters::saveReplicate = false;
bool MerlinParameters::varianceComponents = false;
bool MerlinParameters::writeFrequencyFile = false;

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("General")
      LONG_PARAMETER("error", &MerlinCore::findErrors)
      LONG_PARAMETER("information", &FamilyAnalysis::estimateInformation)
      LONG_PARAMETER("likelihood", &FamilyAnalysis::calcLikelihood)
   LONG_PARAMETER_GROUP("IBD States")
      LONG_PARAMETER("ibd", &FamilyAnalysis::estimateIBD)
      LONG_PARAMETER("kinship", &FamilyAnalysis::estimateKinship)
      LONG_PARAMETER("matrices", &FamilyAnalysis::estimateMatrices)
#ifndef __CHROMOSOME_X__
      LONG_PARAMETER("extended", &FamilyAnalysis::estimateKinship15)
#endif
   LONG_PARAMETER_GROUP("Linkage")
      LONG_PARAMETER("npl", &FamilyAnalysis::nplAll)
      LONG_PARAMETER("pairs", &FamilyAnalysis::nplPairs)
      LONG_PARAMETER("qtl", &FamilyAnalysis::nplQtl)
      LONG_PARAMETER("deviates", &FamilyAnalysis::nplDeviates)
   LONG_PARAMETER_GROUP("VC Linkage")
      LONG_PARAMETER("vc", &MerlinParameters::varianceComponents)
      LONG_PARAMETER("useCovariates", &VarianceComponents::useCovariates)
      LONG_PARAMETER("ascertainment", &VarianceComponents::useProbands)
      LONG_DOUBLEPARAMETER("unlinked", &VarianceComponents::unlinkedFraction)
   LONG_PARAMETER_GROUP("Haplotyping")
      LONG_PARAMETER("best", &FamilyAnalysis::bestHaplotype)
      LONG_SMARTINTPARAMETER("sample", &FamilyAnalysis::sampledHaplotypes)
      LONG_PARAMETER("all", &FamilyAnalysis::allHaplotypes)
      LONG_PARAMETER("founders", &MerlinHaplotype::outputFounders)
      LONG_PARAMETER("horizontal", &MerlinHaplotype::outputHorizontal)
   LONG_PARAMETER_GROUP("Recombination")
      EXCLUSIVE_PARAMETER("zero", &MerlinCore::zeroRecombination)
      EXCLUSIVE_PARAMETER("one", &MerlinCore::oneRecombination)
      EXCLUSIVE_PARAMETER("two", &MerlinCore::twoRecombinations)
      EXCLUSIVE_PARAMETER("three", &MerlinCore::threeRecombinations)
      EXCLUSIVE_PARAMETER("singlepoint", &MerlinCore::twopoint)
   LONG_PARAMETER_GROUP("Positions")
      LONG_INTPARAMETER("steps", &MerlinCore::steps_per_interval)
      LONG_DOUBLEPARAMETER("maxStep", &MerlinCore::maxDistance)
      LONG_DOUBLEPARAMETER("minStep", &MerlinCore::minDistance)
      LONG_DOUBLEPARAMETER("grid", &MerlinCore::grid)
      LONG_DOUBLEPARAMETER("start", &MerlinCore::start)
      LONG_DOUBLEPARAMETER("stop", &MerlinCore::stop)
   LONG_PARAMETER_GROUP("Performance")
      LONG_INTPARAMETER("bits", &MerlinCore::maxBits)
      LONG_INTPARAMETER("megabytes", &MerlinParameters::maxMegabytes)
      LONG_INTPARAMETER("minutes", &MerlinCore::maxMinutes)
      LONG_PARAMETER("trim", &MerlinParameters::trimPedigree)
#ifndef __CHROMOSOME_X__
      LONG_PARAMETER("noCoupleBits", &Mantra::ignoreCoupleSymmetries)
#endif
   LONG_PARAMETER_GROUP("Output")
      LONG_PARAMETER("quiet", &MerlinCore::quietOutput)
      LONG_PARAMETER("markerNames", &MerlinCore::useMarkerNames)
      LONG_PARAMETER("frequencies", &MerlinParameters::writeFrequencyFile)
      LONG_PARAMETER("perFamily", &FamilyAnalysis::perFamily)
      LONG_PARAMETER("pdf", &FamilyAnalysis::writePDF)
   LONG_PARAMETER_GROUP("Simulation")
      LONG_PARAMETER("simulate", &MerlinParameters::simulateNull)
      LONG_PARAMETER("save", &MerlinParameters::saveReplicate)
   LONG_PARAMETER_GROUP("Additional")
      LONG_PARAMETER("simwalk2", &FamilyAnalysis::simwalk2)
      LONG_PARAMETER("swap", &FamilyAnalysis::useSwap)
END_LONG_PARAMETERS();

MerlinParameters::MerlinParameters() :
   pedigreeFile("merlin.ped"),
   dataFile("merlin.dat"),
   mapFile("merlin.map"),
   freqFile(""),
   random_seed(123456),
   alleleFrequencies(FREQ_ALL)
   {
   Add(new StringParameter('d', "Data File", dataFile));
   Add(new StringParameter('p', "Pedigree File", pedigreeFile));
   Add(new StringParameter('x', "Missing Value Code", Pedigree::missing, false));
   Add(new StringParameter('m', "Map File", mapFile));
   Add(new AlleleFrequencyParameter('f', "Allele Frequencies", alleleFrequencies, freqFile));
   Add(new IntParameter('r', "Random Seed", random_seed));
   Add(new LongParameters("Data Analysis Options", longParameters));

   // The -b switch is obsolete and is thus hidden (use --bits instead)
   Add(new HiddenInteger('b', "Maximum Bits", FamilyAnalysis::maxBits));
   // The -i switch is obsolete and is thus hidden (use --steps instead)
   Add(new HiddenInteger('i', "Steps Per Interval", FamilyAnalysis::steps_per_interval));
   }

void MerlinParameters::Read(int argc, char ** argv, int start)
   {
   ParameterList::Read(argc, argv, start);

   globalRandom.Reset(random_seed);

   // The number of analysis steps between markers must be positive
   if (FamilyAnalysis::steps_per_interval < 0)
      FamilyAnalysis::steps_per_interval = 0;

   // Memory usage limit should be positive
   if (maxMegabytes < 0)
      maxMegabytes = 0;

   // The bit limit for analysis should be positive
   if (FamilyAnalysis::maxBits < 0)
      FamilyAnalysis::maxBits = 0;

   // The number of sampled haplotypes should be positive
   if (FamilyAnalysis::sampledHaplotypes < 0)
      FamilyAnalysis::sampledHaplotypes = 0;

   // Limited range of options in two-point mode
   if (FamilyAnalysis::twopoint)
      {
      // No between marker analyses
      FamilyAnalysis::steps_per_interval = 0;

      // No error checking
      FamilyAnalysis::findErrors = false;

      // No haplotyping
      FamilyAnalysis::bestHaplotype = false;
      FamilyAnalysis::sampledHaplotypes = 0;
      FamilyAnalysis::allHaplotypes = false;

      // Positions along chromosome not meaningful
      FamilyAnalysis::useMarkerNames = true;
      }

   // Limited range of options when using simwalk2
   if (FamilyAnalysis::simwalk2)
      {
      // No between marker analyses
      FamilyAnalysis::steps_per_interval = 0;

      // Simwalk2 might want the marker names
      FamilyAnalysis::useMarkerNames = true;

      // Non-parametric statistics are required
      FamilyAnalysis::nplAll = true;
      FamilyAnalysis::nplPairs = true;
      }

   // Checked that proportion of unlinked families is within range
   if (VarianceComponents::unlinkedFraction < 0.0 ||
       VarianceComponents::unlinkedFraction >= 1.0)
      VarianceComponents::unlinkedFraction = 0.0;

   FamilyAnalysis::storeKinship |= varianceComponents;

   BasicTree::maxNodes = (1024 * 1024) / sizeof(TreeNode) * maxMegabytes;

   simulateNull |= saveReplicate;
   }

AlleleFrequencyParameter::AlleleFrequencyParameter
   (char c, const char * desc, int & how, String & file) :
   Parameter(c, desc, &how), status("ALL INDIVIDUALS")
   {
   how = FREQ_ALL;
   filename = &file;
   }

void AlleleFrequencyParameter::Translate(const char * str)
   {
   String value(str);
   int * estimator = (int *) var;

   if (value == "a" || value == "A")
      {
      * estimator = FREQ_ALL;
      * filename = "";
      status = "ALL INDIVIDUALS";
      }
   else if (value == "e" || value == "E")
      {
      * estimator = FREQ_EQUAL;
      * filename = "";
      status = "ASSUMED EQUAL";
      }
   else if (value == "f" || value == "F")
      {
      * estimator = FREQ_FOUNDERS;
      * filename = "";
      status = "FOUNDERS ONLY";
      }
   else
      {
      * estimator = FREQ_ALL;
      * filename = value;
      status = value;
      }
   }

void AlleleFrequencyParameter::Status()
   {
   printf("%*s : %*s (-%c[a|e|f|file])\n", nameCol, description,
          statusCol, (const char *) status, ch);
   }

 
 
