////////////////////////////////////////////////////////////////////// 
// merlin/TreeBasics.h 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday June 11, 2003
// 
 
#ifndef __TREEBASE_H__
#define __TREEBASE_H__

#include "TreeNode.h"
#include "MiniDeflate.h"
#include "IntArray.h"

#include <stdlib.h>
#include <stdio.h>

class BasicTree
   {
   public:
      // Maximum number of nodes to store in memory
      static int maxNodes;
      static int totalNodes;

      // Tree is stored as an array of nodes
      TreeNode * nodes;
      int        nextFree;

      // Depth of tree
      int        bit_count;

      // Total number of nodes allocated
      int        count;

      BasicTree()
         { Initialize(); }

      BasicTree(BasicTree & tree)
         { Initialize(); Copy(tree); }

      ~BasicTree()
         {
         if (nodes != NULL)
            {
            totalNodes -= count;
            free(nodes);
            }
         }

      void Clear()
         { nextFree = 0; }

      int NewNode()
         {
         if (nextFree == count)
            Grow();
         return nextFree++;
         }

      void Grow();

      void Trim(int node = 0);
      void TrimNoMerge(int node = 0);

      int  Copy(int node);
      int  Copy(const BasicTree & source, int node);
      void CarbonCopy(const BasicTree & source);
      void Copy(const BasicTree & source);

      BasicTree & operator = (BasicTree & rhs);

      void Print(int node = 0, int pad = 0);

      void MakeMinimalTree(double value, int bits);
      bool IsEmpty()
         { return nextFree == 0 || nodes[0].type == TREE_NODE_ZERO; }

      static void SetupSwap();
      static void FreeSwap();
      static void CloseSwap();

      // Routines for swapping trees in and out of memory
      void Pack();
      void UnPack();
      void Discard();

   protected:
      // Used for avoiding recursion
      IntArray stack;

   private:
      void Initialize()
         {
         bit_count = nextFree = count = 0;
         nodes = NULL;
         }

      // Routines for managing memory allocation failures
      void MemoryCeiling();
      void OutOfMemory();

      // Global swap file variables
      //
      static char        * skeleton;
      static double      * leaves;
      static FILE        * swap_skeleton;
      static FILE        * swap_leaves;
      static MiniDeflate   zip;

#ifdef _FILE_OFFSET_BITS
#  if _FILE_OFFSET_BITS == 64
#    define __HUGE_SWAP__
#  endif
#endif

#ifndef __HUGE_SWAP__
      // Architecture dependent swap file variables
      //
      static int free_leaves, free_skeleton;
      static int swap_peak;

      // Variables specific to this tree
      //
      int leafCount;
      int offset1, offset2;
#else
      //
      static fpos_t free_leaves, free_skeleton;

      // Variables specific to this tree
      //
      int     leafCount;
      fpos_t offset1, offset2;
#endif
   };

class TreesTooBig
   {
   public:
      int memory_request;

      TreesTooBig(const TreesTooBig & source)
         { memory_request = source.memory_request; }
      TreesTooBig(int request)
         { memory_request = request; }

   };

///////////////////////////////////////////////////////////////
// The SAFE_SET macro is crucial for setting node values
// when MERLIN is compiled by gcc-3.0 and above. SAFE_SET
// ensures that the left side (variable) of an assignment
// expression is  evaluated after the right side (value)
//
// Without it, the following piece of code can raise
// segmentation faults whenever NewNode() changes the
// nodes pointer!
//
//   nodes[i].child[0] = NewNode();
//
// Note that NewNode() and many other tree functions can
// change the value of the nodes pointer!
//

#define SAFE_SET(a,b) {int tmp = b; a = tmp; }

#endif


 
 
